/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Arrays;

import com.fasterxml.jackson.annotation.JsonIgnore;
import lombok.*;
import lombok.experimental.Accessors;
import lombok.experimental.SuperBuilder;
import lombok.extern.slf4j.Slf4j;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.ConvolutionMode;
import org.deeplearning4j.util.ValidationUtils;

/**
 * ConvolutionLayer nIn in the input layer is the number of channels nOut is the number of filters
 * to be used in the net or in other words the channels The builder specifies the filter/kernel
 * size, the stride and padding The pooling layer takes the kernel size
 *
 * <p>Supports multiple dimensions: In 1D CNN, kernel moves in 1 direction. Input and output data of
 * 1D CNN is 2 dimensional. Mostly used on Time-Series data.
 *
 * <p>In 2D CNN, kernel moves in 2 directions. Input and output data of 2D CNN is 3 dimensional.
 * Mostly used on Image data.
 *
 * <p>In 3D CNN, kernel moves in 3 directions. Input and output data of 3D CNN is 4 dimensional.
 * Mostly used on 3D Image data (MRI, CT Scans, Video).
 */
@ToString(callSuper = true)
@NoArgsConstructor
@EqualsAndHashCode(callSuper = true)
@Slf4j
@SuperBuilder
public abstract class AbstractConvolutionLayer extends FeedForwardLayer {
  /** The kernel of this convolution with size in each n-dimensions */
  @Getter private int[] kernelSize;
  /** The stride */
  @Getter private int[] stride;
  /** The padding */
  @Getter private int[] padding;
  /** The dilation */
  @Getter private int[] dilation;
  /** If true (default): include bias parameters in the model. False: no bias. */
  @Builder.Default
  @Getter
  @Accessors(fluent = true)
  @Setter
  private boolean hasBias = true;
  /**
   * Set the convolution mode for the Convolution layer. See {@link ConvolutionMode} for more
   * details Default is {@link ConvolutionMode}.Truncate.
   */
  @Builder.Default @Getter @Setter
  private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;
  /**
   * When using CuDNN and an error is encountered, should fallback to the non-CuDNN implementatation
   * be allowed? If set to false, an exception in CuDNN will be propagated back to the user. If
   * false, the built-in (non-CuDNN) implementation for ConvolutionLayer will be used
   */
  @Getter @Setter @Builder.Default private boolean cudnnAllowFallback = true;

  /** Defaults to "PREFER_FASTEST", but "NO_WORKSPACE" uses less memory. */
  @Getter @Setter @Builder.Default private ConvolutionLayer.AlgoMode cudnnAlgoMode = ConvolutionLayer.AlgoMode.PREFER_FASTEST;

  @Getter @Setter private ConvolutionLayer.FwdAlgo cudnnFwdAlgo;
  @Getter @Setter private ConvolutionLayer.BwdFilterAlgo cudnnBwdFilterAlgo;
  @Getter @Setter private ConvolutionLayer.BwdDataAlgo cudnnBwdDataAlgo;

  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default @Getter @Setter
  private CNN2DFormat convFormat =
          CNN2DFormat.NCHW; // default value for legacy serialization reasons


  /**
   * Number of parameters this layer has a result of its configuration.
   *
   * @return number or parameters
   */
  @Override
  public long numParams() {
    var kern = 1;
    for (int i : getKernelSize()) {
      kern = kern * i;
    }
    return nIn * nOut * kern + (hasBias() ? nOut : 0);
  }

  public abstract static class AbstractConvolutionLayerBuilder<
          C extends AbstractConvolutionLayer, B extends AbstractConvolutionLayerBuilder<C, B>>
      extends FeedForwardLayerBuilder<C, B> {

    public B kernelSize(int @NonNull ... kernelSize) {
      if (this.kernelSize != null) {
        log.warn("You are setting the kernel more than once, last call with override prior calls.");
      }
      this.kernelSize = kernelSize;
      return self();
    }

    public B stride(int @NonNull ... stride) {
      this.stride = stride;
      return self();
    }

    public B padding(int @NonNull ... padding) {
      this.padding = padding;
      return self();
    }

    public B dilation(int @NonNull ... dilation) {
      this.dilation = dilation;
      return self();
    }
  }
}
