/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution1DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;


//TODO: We will eventually want to NOT subclass off of ConvolutionLayer.
//Currently, we just subclass off the ConvolutionLayer and hard code the "width" dimension to 1
/**
 * This approach treats a multivariate time series with L time steps and
 * P variables as an L x 1 x P image (L rows high, 1 column wide, P
 * channels deep). The kernel should be H<L pixels high and W=1 pixels
 * wide.
 *
 * In 1D CNN, kernel moves in 1 direction. The kernel has 2-dimensions.
 * Input and output data of 1D CNN is 2-dimensional. Mostly used on Time-Series data.
 */
@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class Convolution1D extends ConvolutionLayer {
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default
  protected CNN2DFormat dataFormat =
      CNN2DFormat.NCHW; // default value for legacy serialization reasons

  @Builder.Default private RNNFormat rnnDataFormat = RNNFormat.NCW;
  /**
   * Size of the convolution
   *
   * @param kernelSize the length of the kernel
   */
  @Builder.Default private int[] kernelSize = new int[] {1, 1};
  /**
   * Stride for the convolution. Must be > 0
   *
   * @param stride Stride
   */
  @Builder.Default private int[] stride = new int[] {1, 1};
  /**
   * Padding value for the convolution. Not used with {@link
   * org.deeplearning4j.nn.conf.ConvolutionMode#Same}
   *
   * @param padding Padding value
   */
  @Builder.Default private int[] padding = new int[] {0, 0};

  @Builder.Default private int[] dilation = new int[] {1};

  public static Convolution1DBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static Convolution1DBuilder<?, ?> builder(int kernelSize) {
    return innerBuilder()
        .kernelSize(new int[] {kernelSize})
        .stride(new int[] {1, 1})
        .padding(new int[] {0, 0})
        .dilation(new int[] {0});
  }

  public static Convolution1DBuilder<?, ?> builder(int kernelSize, int stride) {
    return innerBuilder()
        .kernelSize(new int[] {kernelSize})
        .stride(new int[] {stride})
        .padding(new int[] {0, 0})
        .dilation(new int[] {0});
  }

  public static Convolution1DBuilder<?, ?> builder(int kernelSize, int stride, int padding) {
    return innerBuilder()
        .kernelSize(new int[] {kernelSize})
        .stride(new int[] {stride})
        .padding(new int[] {padding})
        .dilation(new int[] {0});
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);
    LayerValidation.assertNInNOutSet("Convolution1D", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    org.deeplearning4j.nn.layers.convolution.Convolution1DLayer ret =
        new org.deeplearning4j.nn.layers.convolution.Convolution1DLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D CNN layer (layer index = "
              + layerIndex
              + ", layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }
    InputType.InputTypeRecurrent it = (InputType.InputTypeRecurrent) inputType;
    long inputTsLength = it.getTimeSeriesLength();
    long outLength;
    if (inputTsLength < 0) {
      // Probably: user did InputType.recurrent(x) without specifying sequence length
      outLength = -1;
    } else {
      outLength =
          Convolution1DUtils.getOutputSize(
              inputTsLength,
              kernelSize[0],
              stride[0],
              padding[0],
              getConvolutionMode(),
              dilation[0]);
    }

    return InputType.recurrent(nOut, outLength, rnnDataFormat);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D CNN layer (layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }

    InputType.InputTypeRecurrent r = (InputType.InputTypeRecurrent) inputType;
    if (nIn <= 0 || override) {
      this.nIn = r.getSize();
    }
    if (this.rnnDataFormat == null || override) this.rnnDataFormat = r.getFormat();

    if (this.dataFormat == null || override)
      this.dataFormat = rnnDataFormat == RNNFormat.NCW ? CNN2DFormat.NCHW : CNN2DFormat.NHWC;
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Convolution1D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, rnnDataFormat, getName());
  }

  protected boolean allowCausal() {
    return true;
  }

  private static final class Convolution1DBuilderImpl
      extends Convolution1DBuilder<Convolution1D, Convolution1DBuilderImpl> {
    public Convolution1D build() {
      Convolution1D l = new Convolution1D(this);
      Convolution1DUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding()[0]);
      Convolution1DUtils.validateCnn1DKernelStridePadding(
          l.getKernelSize()[0], l.getStride()[0], l.getPadding()[0]);
      l.initializeConstraints();
      return l;
    }
  }

  public abstract static class Convolution1DBuilder<
          C extends Convolution1D, B extends Convolution1DBuilder<C, B>>
      extends ConvolutionLayerBuilder<C, B> {

    @Override
    public B kernelSize(int @NonNull ... kernelSize) {
      //Todo, we always provide arrays, but only first element is really used
      super.kernelSize(ValidationUtils.validate1NonNegative(new int[]{kernelSize[0]},"kernelSize"));
      return self();
    }

    public B padding(int @NonNull ... padding) {
      //Todo, we always provide arrays, but only first element is really used
      super.padding(ValidationUtils.validate1NonNegative(new int[]{padding[0]}, "padding"));

      return self();
    }
    public B dilation(int @NonNull ... dilation) {
      //Todo, we always provide arrays, but only first element is really used
      super.dilation(ValidationUtils.validate1NonNegative(new int[]{dilation[0]}, "dilation"));
      return self();
    }

    public B stride(int @NonNull ... stride) {
      //Todo, we always provide arrays, but only first element is really used
      super.stride(ValidationUtils.validate1NonNegative(new int[]{stride[0]}, "stride"));
      return self();
    }
  }
}
