/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import lombok.extern.slf4j.Slf4j;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.params.ConvolutionNewParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution1DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

// TODO: We will eventually want to NOT subclass off of ConvolutionLayer.
// Currently, we just subclass off the ConvolutionLayer and hard code the "width" dimension to 1

/**
 * This approach treats a multivariate time series with L time steps and P variables as an L x 1 x P
 * image (L rows high, 1 column wide, P channels deep). The kernel should be H<L pixels high and W=1
 * pixels wide.
 *
 * <p>In 1D CNN, kernel moves in 1 direction. The kernel has 2-dimensions. Input and output data of
 * 1D CNN is 2-dimensional. Mostly used on Time-Series data.
 */
@Data
@Slf4j
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder
public class Convolution1DNew extends AbstractConvolutionLayer {

  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default
  protected CNN2DFormat dataFormat =
      CNN2DFormat.NCHW; // default value for legacy serialization reasons

  @Builder.Default private RNNFormat rnnDataFormat = RNNFormat.NCW;

  @Override
  public ParamInitializer initializer() {
    return ConvolutionNewParamInitializer.getInstance();
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);
    LayerValidation.assertNInNOutSet("Convolution1D", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    /*
       Object ret;
       try {
       ret = lconf.getCanConfigure()
               .getConstructor(LayerConfiguration.class, DataType.class)
               .newInstance(new Object[] { lconf, networkDataType });
       } catch (Exception e) {
         throw new RuntimeException(e);

    */
    org.deeplearning4j.nn.layers.convolution.Convolution1DNewLayer ret =
        new org.deeplearning4j.nn.layers.convolution.Convolution1DNewLayer(lconf, networkDataType);

    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(this);
    return ret;
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D CNN layer (layer index = "
              + layerIndex
              + ", layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }
    InputType.InputTypeRecurrent it = (InputType.InputTypeRecurrent) inputType;
    long inputTsLength = it.getTimeSeriesLength();
    long outLength;
    if (inputTsLength < 0) {
      // Probably: user did InputType.recurrent(x) without specifying sequence length
      outLength = -1;
    } else {
      outLength =
          Convolution1DUtils.getOutputSize(
              inputTsLength,
              getKernelSize()[0],
              getStride()[0],
              getPadding()[0],
              getConvolutionMode(),
              getDilation()[0]);
    }

    return InputType.recurrent(nOut, outLength, rnnDataFormat);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D CNN layer (layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }

    InputType.InputTypeRecurrent r = (InputType.InputTypeRecurrent) inputType;
    if (nIn <= 0 || override) {
      this.nIn = r.getSize();
    }
    if (this.rnnDataFormat == null || override) this.rnnDataFormat = r.getFormat();

    if (this.dataFormat == null || override)
      this.dataFormat = rnnDataFormat == RNNFormat.NCW ? CNN2DFormat.NCHW : CNN2DFormat.NHWC;
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Convolution1D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, rnnDataFormat, getName());
  }

  /**
   * This is a report of the estimated memory consumption for the given layer
   *
   * @param inputType Input type to the layer. Memory consumption is often a function of the input
   *     type
   * @return Memory report for the layer
   */
  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return null;
  }

  protected boolean allowCausal() {
    return true;
  }

  private static final class Convolution1DNewBuilderImpl
      extends Convolution1DNewBuilder<Convolution1DNew, Convolution1DNewBuilderImpl> {
    public Convolution1DNew build() {
      Convolution1DNew l = new Convolution1DNew(this);
      if (l.getDilation() == null) {
        dilation(1, 1);
      }
      if (l.getPadding() == null) {
        padding(0);
      }
      l = new Convolution1DNew(this);

      Convolution1DUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding()[0]);
      Convolution1DUtils.validateCnn1DKernelStridePadding(
          l.getKernelSize()[0], l.getStride()[0], l.getPadding()[0]);
      l.initializeConstraints();
      return l;
    }
  }

  public abstract static class Convolution1DNewBuilder<
          C extends Convolution1DNew, B extends Convolution1DNewBuilder<C, B>>
      extends AbstractConvolutionLayerBuilder<C, B> {
    private int dimensions(Class arrayType) {
      return arrayType.isArray() ? 1 + dimensions(arrayType.getComponentType()) : 0;
    }

    @Override
    public B kernelSize(int @NonNull ... kernel) {
      // Todo, we always provide arrays, but only first element is really used
      if (dimensions(kernel.getClass()) > 1)
        log.warn(
            "Kernel size has '{}' dimensions, only using first dimensions for 1D convolution layer.",
            dimensions(kernel.getClass()));
      super.kernelSize(
          ValidationUtils.validate1NonNegative(new int[] {kernel[0]}, "kernelSize")[0], 1);
      return self();
    }

    public B padding(int @NonNull ... padding) {
      // Todo, we always provide arrays, but only first element is really used
      super.padding(ValidationUtils.validate1NonNegative(new int[] {padding[0]}, "padding"));

      return self();
    }

    public B dilation(int @NonNull ... dilation) {
      // Todo, we always provide arrays, but only first element is really used
      super.dilation(ValidationUtils.validate1NonNegative(new int[] {dilation[0]}, "dilation"));
      return self();
    }

    public B stride(int @NonNull ... stride) {
      // Todo, we always provide arrays, but only first element is really used
      super.stride(ValidationUtils.validate1NonNegative(new int[] {stride[0]}, "stride")[0], 1);
      return self();
    }
  }
}
