/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.*;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.layers.convolution.Convolution3DLayer;
import org.deeplearning4j.nn.params.Convolution3DParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution3DUtils;
import org.deeplearning4j.util.Convolution2DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@NoArgsConstructor
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class Convolution3D extends ConvolutionLayer {

  /**
   * Set dilation size for 3D convolutions in (depth, height, width) order
   *
   * @param dilation kernel size
   */
  private int[] dilation;
  /**
   * Set padding size for 3D convolutions in (depth, height, width) order
   *
   * @param padding kernel size
   */
  private int[] padding;

  private int[] stride;

  @Builder.Default
  private ConvolutionMode mode = ConvolutionMode.Same; // in libnd4j: 0 - same mode, 1 - valid mode

  @Builder.Default
  private DataFormat dataFormat = DataFormat.NCDHW; // in libnd4j: 1 - NCDHW, 0 - NDHWC
  /**
   * Set kernel size for 3D convolutions in (depth, height, width) order
   *
   * @param kernelSize kernel size
   */
  private int[] kernelSize;

  @Builder.Default private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;

  public static Convolution3DBuilder<?, ?> builder() {
    return innerBuilder()
        .kernelSize(new int[] {2, 2, 2})
        .stride(new int[] {1, 1, 1})
        .padding(new int[] {0, 0, 0})
        .dilation(new int[] {1, 1, 1})
        .convolutionDim(3);
  }

  public static Convolution3DBuilder<?, ?> builder(
      int[] kernelSize, int[] stride, int[] padding, int[] dilation) {
    return innerBuilder()
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding)
        .dilation(dilation)
        .convolutionDim(3);
  }

  public static Convolution3DBuilder<?, ?> builder(
          int[] kernelSize, int[] stride, int[] padding) {
    return innerBuilder()
            .kernelSize(kernelSize)
            .stride(stride)
            .padding(padding)
            .convolutionDim(3);
  }



  public static Convolution3DBuilder<?, ?> builder(int[] kernelSize, int[] stride) {
    return innerBuilder().kernelSize(kernelSize).stride(stride).convolutionDim(3);
  }

  public static Convolution3DBuilder<?, ?> builder(int... kernelSize) {
    return innerBuilder().kernelSize(kernelSize).convolutionDim(3);
  }

  /**
   * 3-dimensional convolutional layer configuration nIn in the input layer is the number of
   * channels nOut is the number of filters to be used in the net or in other words the depth The
   * builder specifies the filter/kernel size, the stride and padding The pooling layer takes the
   * kernel size
   */
  public boolean hasBias() {
    return isHasBias();
  }

  @Override
  public Convolution3D clone() {
    Convolution3D clone = (Convolution3D) super.clone();
    if (clone.kernelSize != null) {
      clone.kernelSize = clone.kernelSize.clone();
    }
    if (clone.stride != null) {
      clone.stride = clone.stride.clone();
    }
    if (clone.padding != null) {
      clone.padding = clone.padding.clone();
    }
    if (clone.dilation != null) {
      clone.dilation = clone.dilation.clone();
    }
    return clone;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> iterationListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet("Convolution3D", getName(), layerIndex, getNIn(), getNOut());

    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();

    Convolution3DLayer ret = new Convolution3DLayer(lconf, networkDataType);
    ret.addTrainingListeners(iterationListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return Convolution3DParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
      throw new IllegalStateException(
          "Invalid input for Convolution3D layer (layer name=\""
              + getName()
              + "\"): Expected CNN3D input, got "
              + inputType);
    }
    return InputTypeUtil.getOutputTypeCnn3DLayers(
        inputType,
        dataFormat,
        kernelSize,
        stride,
        padding,
        dilation,
        convolutionMode,
        nOut,
        layerIndex,
        getName(),
        Convolution3DLayer.class);
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Convolution3D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreProcessorForInputTypeCnn3DLayers(inputType, getName());
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
      throw new IllegalStateException(
          "Invalid input for Convolution 3D layer (layer name=\""
              + getName()
              + "\"): Expected CNN3D input, got "
              + inputType);
    }

    if (nIn <= 0 || override) {
      InputType.InputTypeConvolutional3D c = (InputType.InputTypeConvolutional3D) inputType;
      this.nIn = c.getChannels();
    }
  }

  protected boolean allowCausal() {
    // Causal convolution - allowed for 1D only
    return false;
  }

  /**
   * An optional dataFormat: "NDHWC" or "NCDHW". Defaults to "NCDHW".<br>
   * The data format of the input and output data. <br>
   * For "NCDHW" (also known as 'channels first' format), the data storage order is: [batchSize,
   * inputChannels, inputDepth, inputHeight, inputWidth].<br>
   * For "NDHWC" ('channels last' format), the data is stored in the order of: [batchSize,
   * inputDepth, inputHeight, inputWidth, inputChannels].
   */
  public enum DataFormat {
    NCDHW,
    NDHWC
  }

  private static final class Convolution3DBuilderImpl extends Convolution3DBuilder<Convolution3D, Convolution3DBuilderImpl> {
    public Convolution3D build() {
      Convolution3D l = new Convolution3D(this);
      Convolution2DUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding());
      Convolution3DUtils.validateCnn3DKernelStridePadding(l.getKernelSize(), l.getStride(), l.getPadding());
      return l;
    }
  }
  // public Builder(int[] kernelSize, int[] stride, int[] padding, int[] dilation) {
  // sup/er(kernelSize, stride, padding, dilation, 3);
  public static abstract class Convolution3DBuilder<
          C extends Convolution3D, B extends Convolution3DBuilder<C, B>>
      extends ConvolutionLayer.ConvolutionLayerBuilder<C, B> {


    @Override // TODO we can use the parent builder and do not need to redefine the variables.
    // Validation can be done in override function!
    public B stride(int... stride) {
      this.stride = ValidationUtils.validate3NonNegative(stride, "stride");
      return self();
    }

    public B kernelSize(int... kernelSize) {
      this.kernelSize = ValidationUtils.validate3NonNegative(kernelSize, "kernelSize");
      return self();
    }

    @Override
    public B dilation(int... dilation) {
      this.dilation = ValidationUtils.validate3NonNegative(dilation, "dilation");
      return self();
    }

    @Override
    public B padding(int... padding) {
      this.padding = ValidationUtils.validate3NonNegative(padding, "padding");
      return self();
    }


  }
}
