/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.ConvolutionMode;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.layers.convolution.Deconvolution3DLayer;
import org.deeplearning4j.nn.params.Deconvolution3DParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/**
 * Deconvolution3D layer nIn in the input layer is the number of channels nOut is the number of
 * filters to be used in the net or in other words the channels The builder specifies the
 * filter/kernel size, the stride and padding The pooling layer takes the kernel size
 */
@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder(builderMethodName = "innerBuilder")
public class Deconvolution3D extends ConvolutionLayer {
  /**
   * Set the convolution mode for the Convolution layer. See {@link ConvolutionMode} for more
   * details
   *
   * @param convolutionMode Convolution mode for layer
   */
  @lombok.Builder.Default
  private Convolution3D.DataFormat dataFormat =
      Convolution3D.DataFormat.NCDHW; // in libnd4j: 1 - NCDHW, 0 - NDHWC

  public static Deconvolution3DBuilder<?, ?> builder() {
    return innerBuilder()
        .kernelSize(new int[] {2, 2, 2})
        .stride(new int[] {1, 1, 1})
        .padding(new int[] {0, 0, 0})
        .dilation(new int[] {1, 1, 1})
        .convolutionDim(3);
  }

  protected boolean allowCausal() {
    // Causal convolution - allowed for 1D only
    return false;
  }

  public boolean hasBias() {
    return isHasBias();
  }

  @Override
  public Deconvolution3D clone() {
    Deconvolution3D clone = (Deconvolution3D) super.clone();
    if (clone.getKernelSize() != null) {
      clone.setKernelSize(clone.getKernelSize().clone());
    }
    if (clone.getStride() != null) {
      clone.setStride(clone.getStride().clone());
    }
    if (clone.getPadding() != null) {
      clone.setPadding(clone.getPadding().clone());
    }
    return clone;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet("Deconvolution2D", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    Deconvolution3DLayer ret = new Deconvolution3DLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return Deconvolution3DParamInitializer.getInstance();
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Deconvolution3D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreProcessorForInputTypeCnn3DLayers(inputType, getName());
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
      throw new IllegalStateException(
          "Invalid input for Deconvolution 3D layer (layer name=\""
              + getName()
              + "\"): Expected CNN3D input, got "
              + inputType);
    }

    if (nIn <= 0 || override) {
      InputType.InputTypeConvolutional3D c = (InputType.InputTypeConvolutional3D) inputType;
      this.nIn = c.getChannels();
    }
  }

  // private int[] kernelSize;
  // private int[] stride;
  // private int[] padding;
  // private int[] dilation;

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
      throw new IllegalStateException(
          "Invalid input for Deconvolution layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    return InputTypeUtil.getOutputTypeDeconv3dLayer(
        inputType,
        getKernelSize(),
        getStride(),
        getPadding(),
        getDilation(),
        getConvolutionMode(),
        dataFormat,
        nOut,
        layerIndex,
        getName(),
        Deconvolution3DLayer.class);
  }

  public abstract static class Deconvolution3DBuilder<
          C extends Deconvolution3D, B extends Deconvolution3DBuilder<C, B>>
      extends ConvolutionLayerBuilder<C, B> {}

  private static final class Deconvolution3DBuilderImpl
      extends Deconvolution3DBuilder<Deconvolution3D, Deconvolution3DBuilderImpl> {
    public Deconvolution3D build() {
      Deconvolution3D l = new Deconvolution3D(this);
      l.initializeConstraints();
      return l;
    }
  }
}
