/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.DefaultParamInitializer;
import org.deeplearning4j.nn.weights.WeightInitXavier;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/** Dense Layer Uses WeightInitXavier as default */
@Data
@NoArgsConstructor
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder
public class DenseLayer extends FeedForwardLayer {

  /** If true (default = false): enable layer normalization on this layer */
  @lombok.Builder.Default private boolean hasLayerNorm = false;

  @lombok.Builder.Default private boolean hasBias = true;

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {

    LayerValidation.assertNInNOutSet(
        "DenseLayerConfiguration", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    lconf.setNetConfiguration(conf);
    runInheritance();

    org.deeplearning4j.nn.layers.feedforward.dense.DenseLayer ret =
        new org.deeplearning4j.nn.layers.feedforward.dense.DenseLayer(lconf, networkDataType);

    if (getWeightInit() == null) setWeightInit(new WeightInitXavier());
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return DefaultParamInitializer.getInstance();
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    InputType outputType = getOutputType(-1, inputType);

    val numParams = initializer().numParams(this);
    val updaterStateSize = (int) getUpdater().stateSize(numParams);

    int trainSizeFixed = 0;
    int trainSizeVariable = 0;
    if (getDropOut() != null) {
      if (false) {
        // TODO drop connect
        // Dup the weights... note that this does NOT depend on the minibatch size...
        trainSizeVariable += 0; // TODO
      } else {
        // Assume we dup the input
        trainSizeVariable += inputType.arrayElementsPerExample();
      }
    }

    // Also, during backprop: we do a preOut call -> gives us activations size equal to the output
    // size
    // which is modified in-place by activation function backprop
    // then we have 'epsilonNext' which is equivalent to input size
    trainSizeVariable += outputType.arrayElementsPerExample();

    return new LayerMemoryReport.Builder(name, DenseLayer.class, inputType, outputType)
        .standardMemory(numParams, updaterStateSize)
        .workingMemory(
            0,
            0,
            trainSizeFixed,
            trainSizeVariable) // No additional memory (beyond activations) for inference
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS,
            MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching in DenseLayerConfiguration
        .build();
  }

  public abstract static class DenseLayerBuilder<
          C extends DenseLayer, B extends DenseLayerBuilder<C, B>>
      extends FeedForwardLayerBuilder<C, B> {}

  private static final class DenseLayerBuilderImpl
      extends DenseLayerBuilder<DenseLayer, DenseLayerBuilderImpl> {
    public DenseLayer build() {
      DenseLayer l = new DenseLayer(this);
      l.initializeConstraints();
      return l;
    }
  }
}
