/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.*;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.layers.convolution.DepthwiseConvolution2DLayer;
import org.deeplearning4j.nn.params.DepthwiseConvolutionParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution2DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder
public class DepthwiseConvolution2D extends ConvolutionLayer {
  /**
   * Set channels multiplier for depth-wise convolution
   *
   * @param depthMultiplier integer value, for each input map we get depthMultiplier outputs in
   *     channels-wise step.
   * @return Builder
   */
  @Builder.Default protected int depthMultiplier = 1;
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default
  protected CNN2DFormat dataFormat =
      CNN2DFormat.NCHW; // default value for legacy serialization reasons
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default protected CNN2DFormat cnn2DFormat = CNN2DFormat.NCHW;

  protected boolean allowCausal() {
    // Causal convolution - allowed for 1D only
    return false;
  }

  @Override
  public DepthwiseConvolution2D clone() {
    DepthwiseConvolution2D clone = (DepthwiseConvolution2D) super.clone();
    clone.depthMultiplier = depthMultiplier;
    return clone;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet(
        "DepthwiseConvolution2D", getName(), layerIndex, getNIn(), getNOut());

    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    DepthwiseConvolution2DLayer ret = new DepthwiseConvolution2DLayer(lconf, networkDataType);

    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);

    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return DepthwiseConvolutionParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for  depth-wise convolution layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    return InputTypeUtil.getOutputTypeCnnLayers(
        inputType,
        getKernelSize(),
        getStride(),
        getPadding(),
        getDilation(),
        getConvolutionMode(),
        nOut,
        layerIndex,
        getName(),
        dataFormat,
        DepthwiseConvolution2DLayer.class);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    super.setNIn(inputType, override);

    if (nOut == 0 || override) {
      nOut = this.nIn * this.depthMultiplier;
    }
    this.dataFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
  }

  private static final class DepthwiseConvolution2DBuilderImpl extends DepthwiseConvolution2DBuilder<DepthwiseConvolution2D, DepthwiseConvolution2DBuilderImpl> {
    public DepthwiseConvolution2D build() {
      DepthwiseConvolution2D l = new DepthwiseConvolution2D(this);
      Preconditions.checkState(
              l.getDepthMultiplier() > 0,
              "Depth multiplier must be > 0,  got %s",
              l.getDepthMultiplier());

      Convolution2DUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding());
      Convolution2DUtils.validateCnnKernelStridePadding(
              l.getKernelSize(), l.getStride(), l.getPadding());
      l.initializeConstraints();
      return l;
    }
  }
  public abstract static class DepthwiseConvolution2DBuilder<
          C extends DepthwiseConvolution2D, B extends DepthwiseConvolution2DBuilder<C, B>>
      extends ConvolutionLayerBuilder<C, B> {


    @Override
    public B kernelSize(int... kernelSize) {
      super.kernelSize(ValidationUtils.validate2NonNegative(kernelSize, false, "kernelSize"));
      return self();
    }

    @Override
    public B stride(int... stride) {
      super.stride(ValidationUtils.validate2NonNegative(stride, false, "stride"));
      return self();
    }

    @Override
    public B padding(int... padding) {
      super.padding(ValidationUtils.validate2NonNegative(padding, false, "padding"));
      return self();
    }

    @Override
    public B dilation(int... dilation) {
      super.dilation(ValidationUtils.validate2NonNegative(dilation, false, "dilation"));
      return self();
    }
  }
}
