/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.regularization.Regularization;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder(builderMethodName = "innerBuilder")
public class LocalResponseNormalization extends LayerConfiguration {

  /** Number of adjacent kernel maps to use when doing LRN. default: 5 */
  protected @lombok.Builder.Default double n = 5; // # adjacent kernal maps
  /** LRN scaling constant k. Default: 2 */
  protected @lombok.Builder.Default double k = 2; // constant (e.g. scale)
  /** Scaling constant beta. Default: 0.75 */
  protected @lombok.Builder.Default double beta = 0.75; // decay rate
  /** LRN scaling constant alpha. Default: 1e-4 */
  protected @lombok.Builder.Default double alpha = 1e-4; // decay rate

  // Defaults as per http://www.cs.toronto.edu/~fritz/absps/imagenet.pdf
  // Set defaults here as well as in builder, in case users use no-arg constructor instead of
  // builder
  // defaults based on AlexNet model
  /**
   * When using CuDNN and an error is encountered, should fallback to the non-CuDNN implementatation
   * be allowed? If set to false, an exception in CuDNN will be propagated back to the user. If
   * false, the built-in (non-CuDNN) implementation for BatchNormalization will be used
   */
  protected @lombok.Builder.Default boolean cudnnAllowFallback = true;

  protected @lombok.Builder.Default CNN2DFormat dataFormat = CNN2DFormat.NCHW;

  public static LocalResponseNormalizationBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static LocalResponseNormalizationBuilder<?, ?> builder(
      double k, double n, double alpha, double beta) {
    return innerBuilder().k(k).n(n).alpha(alpha).beta(beta);
  }

  public static LocalResponseNormalizationBuilder<?, ?> builder(
      double k, double alpha, double beta) {
    return innerBuilder().k(k).alpha(alpha).beta(beta);
  }

  @Override
  public LocalResponseNormalization clone() {
    LocalResponseNormalization clone = (LocalResponseNormalization) super.clone();
    return clone;
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.normalization.LocalResponseNormalization ret =
        new org.deeplearning4j.nn.layers.normalization.LocalResponseNormalization(
            lconf, networkDataType);

    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input type for LRN layer (layer index = "
              + layerIndex
              + ", layer name = \""
              + getName()
              + "\"): Expected input of type CNN, got "
              + inputType);
    }
    return inputType;
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    Preconditions.checkState(
        inputType.getType() == InputType.Type.CNN,
        "Only CNN input types can be used with LocalResponseNormalisation, got %s",
        inputType);
    this.dataFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input type for LRN layer (layer name = \"" + getName() + "\"): null");
    }

    return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    return null;
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    return false; // No params in LRN
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    val actElementsPerEx = inputType.arrayElementsPerExample();

    // Forward pass: 3x input size as working memory, in addition to output activations
    // Backward pass: 2x input size as working memory, in addition to epsilons

    return new LayerMemoryReport.Builder(name, DenseLayer.class, inputType, inputType)
        .standardMemory(0, 0)
        .workingMemory(0, 2 * actElementsPerEx, 0, 3 * actElementsPerEx)
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS,
            MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching in DenseLayerConfiguration
        .build();
  }

  public static abstract class LocalResponseNormalizationBuilder<
          C extends LocalResponseNormalization, B extends LocalResponseNormalizationBuilder<C, B>>
      extends LayerConfigurationBuilder<C, B> {
    public B helperAllowFallback(boolean b) {
      this.cudnnAllowFallback$value = b;
      this.cudnnAllowFallback$set = true;
      return self();
    }
  }
}
