/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.ToString;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.lossfunctions.ILossFunction;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder
public class LossLayer extends FeedForwardLayer {

  protected ILossFunction lossFunction;

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.LossLayer ret =
        new org.deeplearning4j.nn.layers.LossLayer(lconf, networkDataType);

    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    throw new UnsupportedOperationException("LossLayer does not contain parameters");
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    // During inference and training: dup the input array. But, this counts as *activations* not
    // working memory
    return new LayerMemoryReport.Builder(name, LossLayer.class, inputType, inputType)
        .standardMemory(0, 0) // No params
        .workingMemory(0, 0, 0, 0)
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  public static abstract class LossLayerBuilder<
          C extends LossLayer, B extends LossLayerBuilder<C, B>>
      extends FeedForwardLayerBuilder<C, B> {
    @Override
    public B nIn(long nIn) {
      throw new UnsupportedOperationException(
          "Ths layer has no parameters, thus nIn will always equal nOut.");
    }

    @Override
    public B nOut(long nOut) {
      throw new UnsupportedOperationException(
          "Ths layer has no parameters, thus nIn will always equal nOut.");
    }
  }
}
