/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.ToString;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.params.DefaultParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.lossfunctions.LossFunctions;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class RnnOutputLayer extends BaseOutputLayer {

  /**
   * @param rnnDataFormat Data format expected by the layer. NCW = [miniBatchSize, size,
   *     timeSeriesLength], NWC = [miniBatchSize, timeSeriesLength, size]. Defaults to NCW.
   */
  private RNNFormat dataFormat;

  public static RnnOutputLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }

  /**
   * @param lossFn Loss function for the output layer
   */
  public static RnnOutputLayerBuilder<?, ?> builder(LossFunctions.LossFunction lossFn) {
    return innerBuilder().lossFunction(lossFn);
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet("RnnOutputLayer", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);

    org.deeplearning4j.nn.layers.recurrent.RnnOutputLayer ret =
        new org.deeplearning4j.nn.layers.recurrent.RnnOutputLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return DefaultParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input type for RnnOutputLayer (layer index = "
              + layerIndex
              + ", layer name=\""
              + getName()
              + "\"): Expected RNN input, got "
              + inputType);
    }
    InputType.InputTypeRecurrent itr = (InputType.InputTypeRecurrent) inputType;

    return InputType.recurrent(nOut, itr.getTimeSeriesLength(), itr.getFormat());
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input type for RnnOutputLayer (layer name=\""
              + getName()
              + "\"): Expected RNN input, got "
              + inputType);
    }

    InputType.InputTypeRecurrent r = (InputType.InputTypeRecurrent) inputType;
    if (dataFormat == null || override) {
      this.dataFormat = r.getFormat();
    }

    if (nIn <= 0 || override) {
      this.nIn = r.getSize();
    }
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, dataFormat, getName());
  }

  public abstract static class RnnOutputLayerBuilder<
          C extends RnnOutputLayer, B extends RnnOutputLayerBuilder<C, B>>
      extends BaseOutputLayerBuilder<C, B> {}

  private static final class RnnOutputLayerBuilderImpl
      extends RnnOutputLayerBuilder<RnnOutputLayer, RnnOutputLayerBuilderImpl> {
    public RnnOutputLayer build() {
      RnnOutputLayer l = new RnnOutputLayer(this);
      l.initializeConstraints();
      return l;
    }
  }
}
