/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder(builderMethodName = "innerBuilder")
public class SpaceToBatchLayer extends NoParamLayer {

  /**
   * Block size for SpaceToBatch layer. Should be a length 2 array for the height and width
   * dimensions
   */
  protected int[] blockSize;
  /** A 2d array, with format [[padTop, padBottom], [padLeft, padRight]] */
  @Builder.Default protected int[][] padding = new int[][] {{0, 0}, {0, 0}};
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default protected CNN2DFormat dataFormat = CNN2DFormat.NCHW;

  public static SpaceToBatchLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }
  // TODO: throw error when block and padding dims don't match

  /**
   * @param blocks Block size for SpaceToBatch layer. Should be a length 2 array for the height and
   *     width dimensions
   */
  public static SpaceToBatchLayerBuilder<?, ?> builder(int[] blocks) {
    return innerBuilder().blockSize(blocks);
  }

  /**
   * @param blocks Block size for SpaceToBatch layer. Should be a length 2 array for the height and
   *     width dimensions
   * @param padding Padding - should be a 2d array, with format [[padTop, padBottom], [padLeft,
   *     padRight]]
   */
  public static SpaceToBatchLayerBuilder<?, ?> builder(int[] blocks, int[][] padding) {
    return innerBuilder().blockSize(blocks).padding(padding);
  }

  @Override
  public SpaceToBatchLayer clone() {
    return (SpaceToBatchLayer) super.clone();
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);

    org.deeplearning4j.nn.layers.convolution.SpaceToBatch ret =
        new org.deeplearning4j.nn.layers.convolution.SpaceToBatch(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional) inputType;
    InputType.InputTypeConvolutional outputType =
        (InputType.InputTypeConvolutional) getOutputType(-1, inputType);

    return new LayerMemoryReport.Builder(name, SpaceToBatchLayer.class, inputType, outputType)
        .standardMemory(0, 0) // No params
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Subsampling layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }
    InputType.InputTypeConvolutional i = (InputType.InputTypeConvolutional) inputType;
    return InputType.convolutional(
        (i.getHeight() + padding[0][0] + padding[0][1]) / blockSize[0],
        (i.getWidth() + padding[1][0] + padding[1][1]) / blockSize[1],
        i.getChannels(),
        i.getFormat());
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    Preconditions.checkState(
        inputType.getType() == InputType.Type.CNN,
        "Only CNN input types can be used with SpaceToBatchLayer, got %s",
        inputType);
    this.dataFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for space to batch layer (layer name=\""
              + getName()
              + "\"): input is null");
    }
    return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    throw new UnsupportedOperationException("SpaceToBatchLayer does not contain parameters");
  }

  public static abstract class SpaceToBatchLayerBuilder<
          C extends SpaceToBatchLayer, B extends SpaceToBatchLayerBuilder<C, B>>
      extends NoParamLayerBuilder<C, B> {
    /**
     * @param blocks Block size for SpaceToBatch layer. Should be a length 2 array for the height
     *     and width dimensions
     * @return
     */
    public B blockSize(int... blocks) {
      this.blockSize = ValidationUtils.validate2NonNegative(blocks, false, "blocks");
      return self();
    }

    /**
     * @param padding Padding - should be a 2d array, with format [[padTop, padBottom], [padLeft,
     *     padRight]]
     * @return
     */
    public B padding(int[][] padding) {
      this.padding$value = ValidationUtils.validate2x2NonNegative(padding, "padding");
      this.padding$set = true;
      return self();
    }
  }
}
