/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.ConvolutionMode;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution3DUtils;
import org.deeplearning4j.util.Convolution2DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.exception.ND4JArraySizeException;
import org.nd4j.linalg.learning.regularization.Regularization;

@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder(builderMethodName = "innerBuilder")
public class Subsampling3DLayer extends NoParamLayer {

  @Builder.Default protected ConvolutionMode convolutionMode = ConvolutionMode.Truncate;
  protected org.deeplearning4j.nn.conf.layers.PoolingType poolingType;
  /**
   * Kernel size
   *
   * @param kernelSize kernel size in height and width dimensions
   */
  protected int[] kernelSize;
  /**
   * Stride
   *
   * @param stride stride in height and width dimensions
   */
  protected int[] stride;
  /**
   * Padding
   *
   * @param padding padding in the height and width dimensions
   */
  protected int[] padding;

  protected int[] dilation;
  @Builder.Default protected boolean cudnnAllowFallback = true;
  /**
   * The data format for input and output activations.<br>
   * NCDHW: activations (in/out) should have shape [minibatch, channels, depth, height, width]<br>
   * NDHWC: activations (in/out) should have shape [minibatch, depth, height, width, channels]<br>
   */
  @Builder.Default
  protected Convolution3D.DataFormat dataFormat =
      Convolution3D.DataFormat.NCDHW; // Default for 1.0.0-beta3 and earlier (before config added)

  public static Subsampling3DLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(org.deeplearning4j.nn.conf.layers.PoolingType poolingType) {
    return innerBuilder()
            .poolingType(poolingType);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(
      org.deeplearning4j.nn.conf.layers.PoolingType poolingType,
      int[] kernelSize,
      int[] stride,
      int[] padding) {
    return innerBuilder()
        .poolingType(poolingType)
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(
      int[] kernelSize, int[] stride, int[] padding) {
    return innerBuilder().kernelSize(kernelSize).stride(stride).padding(padding);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(
      PoolingType poolingType, int[] kernelSize, int[] stride, int[] padding) {
    return innerBuilder()
        .poolingType(poolingType.toPoolingType())
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(PoolingType poolingType, int[] kernelSize) {
    return innerBuilder().poolingType(poolingType.toPoolingType()).kernelSize(kernelSize);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(
      org.deeplearning4j.nn.conf.layers.PoolingType poolingType, int[] kernelSize) {
    return innerBuilder().poolingType(poolingType).kernelSize(kernelSize);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(int[] kernelSize, int[] stride) {
    return innerBuilder().kernelSize(kernelSize).stride(stride);
  }

  public static Subsampling3DLayerBuilder<?, ?> builder(int... kernelSize) {
    return innerBuilder().kernelSize(kernelSize);
  }

  @Override
  public Subsampling3DLayer clone() {
    Subsampling3DLayer clone = (Subsampling3DLayer) super.clone();

    if (clone.kernelSize != null) {
      clone.kernelSize = clone.kernelSize.clone();
    }
    if (clone.stride != null) {
      clone.stride = clone.stride.clone();
    }
    if (clone.padding != null) {
      clone.padding = clone.padding.clone();
    }
    if (clone.dilation != null) {
      clone.dilation = clone.dilation.clone();
    }
    return clone;
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> iterationListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.convolution.subsampling.Subsampling3DLayer ret =
        new org.deeplearning4j.nn.layers.convolution.subsampling.Subsampling3DLayer(
            lconf, networkDataType);
    ret.addTrainingListeners(iterationListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
      throw new IllegalStateException(
          "Invalid input for Subsampling 3D layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    long inChannels = ((InputType.InputTypeConvolutional3D) inputType).getChannels();
    if (inChannels > Integer.MAX_VALUE) throw new ND4JArraySizeException();
    return InputTypeUtil.getOutputTypeCnn3DLayers(
        inputType,
        dataFormat,
        kernelSize,
        stride,
        padding,
        new int[] {1, 1, 1}, // no dilation
        convolutionMode,
        (int) inChannels,
        layerIndex,
        getName(),
        Subsampling3DLayer.class);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // No op: subsampling layer doesn't have nIn value
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Subsampling 3D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreProcessorForInputTypeCnn3DLayers(inputType, getName());
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    // Not applicable
    return null;
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    throw new UnsupportedOperationException("SubsamplingLayer does not contain parameters");
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    InputType.InputTypeConvolutional3D c = (InputType.InputTypeConvolutional3D) inputType;
    InputType.InputTypeConvolutional3D outputType =
        (InputType.InputTypeConvolutional3D) getOutputType(-1, inputType);
    val actElementsPerEx = outputType.arrayElementsPerExample();

    // During forward pass: im2col array + reduce. Reduce is counted as activations, so only im2col
    // is working mem
    val im2colSizePerEx =
        c.getChannels()
            * outputType.getHeight()
            * outputType.getWidth()
            * outputType.getDepth()
            * kernelSize[0]
            * kernelSize[1];

    // Current implementation does NOT cache im2col etc... which means: it's recalculated on each
    // backward pass
    long trainingWorkingSizePerEx = im2colSizePerEx;
    if (getDropOut() != null) {
      // Dup on the input before dropout, but only for training
      trainingWorkingSizePerEx += inputType.arrayElementsPerExample();
    }

    return new LayerMemoryReport.Builder(name, Subsampling3DLayer.class, inputType, outputType)
        .standardMemory(0, 0) // No params
        .workingMemory(0, im2colSizePerEx, 0, trainingWorkingSizePerEx)
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  public enum PoolingType {
    MAX,
    AVG;

    public org.deeplearning4j.nn.conf.layers.PoolingType toPoolingType() {
      switch (this) {
        case MAX:
          return org.deeplearning4j.nn.conf.layers.PoolingType.MAX;
        case AVG:
          return org.deeplearning4j.nn.conf.layers.PoolingType.AVG;
      }
      throw new UnsupportedOperationException("Unknown/not supported pooling type: " + this);
    }
  }

  public static abstract class Subsampling3DLayerBuilder<
          C extends Subsampling3DLayer, B extends Subsampling3DLayerBuilder<C, B>>
      extends NoParamLayerBuilder<C, B> {

    public B kernelSize(int... kernelSize) {
      this.kernelSize = ValidationUtils.validate3NonNegative(kernelSize, "kernelSize");
      return self();
    }

    public B stride(int... stride) {
      this.stride = ValidationUtils.validate3NonNegative(stride, "stride");
      return self();
    }

    public B padding(int... padding) {
      this.padding = ValidationUtils.validate3NonNegative(padding, "padding");
      return self();
    }

    public B dilation(int... padding) {
      this.dilation = ValidationUtils.validate3NonNegative(padding, "dilation");
      return self();
    }


  }

  private static final class Subsampling3DLayerBuilderImpl extends Subsampling3DLayerBuilder<Subsampling3DLayer, Subsampling3DLayerBuilderImpl> {
    public Subsampling3DLayer build() {
      Subsampling3DLayer l = new Subsampling3DLayer(this);
      if (l.getKernelSize().length != 3) {
        throw new IllegalArgumentException("Kernel size must be length 3");
      }

      if (l.getStride().length != 3) {
        throw new IllegalArgumentException("Invalid stride, must be length 3");
      }

      Convolution2DUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding());
      Convolution3DUtils.validateCnn3DKernelStridePadding(l.getKernelSize(), l.getStride(), l.getPadding());
      return l;
    }
  }
}
