/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@EqualsAndHashCode(callSuper = true)
@Jacksonized @SuperBuilder(builderMethodName = "innerBuilder")
public class ZeroPadding1DLayer extends NoParamLayer {
  /** Padding value for left and right. Must be length 2 array */
  @Builder.Default private int[] padding = new int[] {0, 0}; // Padding: left, right

  /**
   * @param padding Padding for both the left and right
   */
  public static ZeroPadding1DLayerBuilder<?, ?> builder(int padding) {
    return innerBuilder().padding(padding, padding);
  }

  public static ZeroPadding1DLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }
  /**
   * @param paddingLeft Padding value for left
   * @param paddingRight Padding value for right
   */
  public static ZeroPadding1DLayerBuilder<?, ?> builder(int paddingLeft, int paddingRight) {
    return innerBuilder().padding(paddingLeft, paddingRight);
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.convolution.ZeroPadding1DLayer ret =
        new org.deeplearning4j.nn.layers.convolution.ZeroPadding1DLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D CNN layer (layer index = "
              + layerIndex
              + ", layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }
    InputType.InputTypeRecurrent recurrent = (InputType.InputTypeRecurrent) inputType;
    return InputType.recurrent(
        recurrent.getSize(), recurrent.getTimeSeriesLength() + padding[0] + padding[1]);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // No op
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for ZeroPadding1DLayer layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, RNNFormat.NCW, getName());
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    throw new UnsupportedOperationException("ZeroPaddingLayer does not contain parameters");
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    InputType outputType = getOutputType(-1, inputType);

    return new LayerMemoryReport.Builder(name, ZeroPaddingLayer.class, inputType, outputType)
        .standardMemory(0, 0) // No params
        .workingMemory(0, 0, MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS)
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  public static abstract class ZeroPadding1DLayerBuilder<
          C extends ZeroPadding1DLayer, B extends ZeroPadding1DLayerBuilder<C, B>>
      extends NoParamLayerBuilder<C, B> {
    public B padding(int... padding) {
      this.padding$value = ValidationUtils.validate2NonNegative(padding, false, "padding");
      this.padding$set = true;
      return self();
    }
  }
}
