/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.convolutional;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.InputTypeUtil;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.layers.NoParamLayer;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.layers.convolution.Cropping1DLayer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/** Amount of cropping to apply to both the top and the bottom of the input activations */
@Data
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class Cropping1D extends NoParamLayer {
  /**
   * Cropping amount for top/bottom (in that order). Must be length 1 or 2 array.
   *
   * @param cropping Cropping amount for top/bottom (in that order). Must be length 1 or 2 array.
   */
  @Builder.Default private int[] cropping = new int[] {0, 0};

  public static Cropping1DBuilder<?, ?> builder() {
    return innerBuilder();
  }
  /**
   * @param cropTopBottom Amount of cropping to apply to both the top and the bottom of the input
   *     activations
   */
  public static Cropping1DBuilder<?, ?> builder(int cropTopBottom) {
    return innerBuilder().cropping(cropTopBottom, cropTopBottom);
  }
  /**
   * @param cropTop Amount of cropping to apply to the top of the input activations
   * @param cropBottom Amount of cropping to apply to the bottom of the input activations
   */
  public static Cropping1DBuilder<?, ?> builder(int cropTop, int cropBottom) {
    return innerBuilder().cropping(cropTop, cropBottom);
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    Cropping1DLayer ret = new Cropping1DLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for 1D Cropping layer (layer index = "
              + layerIndex
              + ", layer name = \""
              + getName()
              + "\"): expect RNN input type with size > 0. Got: "
              + inputType);
    }
    InputType.InputTypeRecurrent cnn1d = (InputType.InputTypeRecurrent) inputType;
    val length = cnn1d.getTimeSeriesLength();
    val outLength = length - cropping[0] - cropping[1];
    return InputType.recurrent(cnn1d.getSize(), outLength);
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    Preconditions.checkArgument(
        inputType != null,
        "Invalid input for Cropping1D layer (layer name=\"" + getName() + "\"): InputType is null");
    return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return null;
  }

  public static abstract class Cropping1DBuilder<
          C extends Cropping1D, B extends Cropping1DBuilder<C, B>>
      extends NoParamLayerBuilder<C, B> {
    public B cropping(int... cropping) {
      this.cropping$value = ValidationUtils.validate2NonNegative(cropping, true, "cropping");
      this.cropping$set = true;
      return self();
    }
  }
}
