/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.convolutional;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.InputTypeUtil;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.layers.NoParamLayer;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.layers.convolution.Cropping2DLayer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution2DUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class Cropping2D extends NoParamLayer {
  /** Cropping amount for top/bottom/left/right (in that order). A length 4 array. */
  @Builder.Default private int[] cropping = new int[] {0, 0, 0, 0};
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default private CNN2DFormat dataFormat = CNN2DFormat.NCHW;

  public static Cropping2DBuilder<?, ?> builder() {
    return innerBuilder();
  }

  /**
   * @param cropping Cropping amount for top/bottom/left/right (in that order). Must be length 4
   *     array.
   */
  public static Cropping2DBuilder<?, ?> builder(int... cropping) {
    return innerBuilder().cropping(cropping);
  }

  /**
   * @param cropTopBottom Amount of cropping to apply to both the top and the bottom of the input
   *     activations
   * @param cropLeftRight Amount of cropping to apply to both the left and the right of the input
   *     activations
   */
  public static Cropping2DBuilder<?, ?> builder(int cropTopBottom, int cropLeftRight) {
    return innerBuilder().cropping(cropTopBottom, cropTopBottom, cropLeftRight, cropLeftRight);
  }

  /**
   * @param cropTop Amount of cropping to apply to the top of the input activations
   * @param cropBottom Amount of cropping to apply to the bottom of the input activations
   * @param cropLeft Amount of cropping to apply to the left of the input activations
   * @param cropRight Amount of cropping to apply to the right of the input activations
   */
  public static Cropping2DBuilder<?, ?> builder(
      int cropTop, int cropBottom, int cropLeft, int cropRight) {
    return innerBuilder().cropping(cropTop, cropBottom, cropLeft, cropRight);
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();

    Cropping2DLayer ret = new Cropping2DLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    int[] hwd = Convolution2DUtils.getHWDFromInputType(inputType);
    int outH = hwd[0] - cropping[0] - cropping[1];
    int outW = hwd[1] - cropping[2] - cropping[3];

    InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional) inputType;

    return InputType.convolutional(outH, outW, hwd[2], c.getFormat());
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    Preconditions.checkArgument(
        inputType != null,
        "Invalid input for Cropping2D layer (layer name=\"" + getName() + "\"): InputType is null");
    return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return null;
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    this.dataFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
  }

  public static abstract class Cropping2DBuilder<
          C extends Cropping2D, B extends Cropping2DBuilder<C, B>>
      extends NoParamLayerBuilder<C, B> {
    public B cropping(int... cropping) {
      this.cropping$value = ValidationUtils.validate4NonNegative(cropping, "cropping");
      this.cropping$set = true;
      return self();
    }
  }
}
