/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.convolutional;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.InputTypeUtil;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.layers.NoParamLayer;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.layers.convolution.Cropping3DLayer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@NoArgsConstructor
@EqualsAndHashCode(callSuper = true)
@Jacksonized
@SuperBuilder(builderMethodName = "innerBuilder")
public class Cropping3D extends NoParamLayer {
@Builder.Default
    private int[] cropping = new int[] {0, 0, 0, 0, 0, 0};;

    /**
     * @param cropDepth Amount of cropping to apply to both depth boundaries of the input activations
     * @param cropHeight Amount of cropping to apply to both height boundaries of the input activations
     * @param cropWidth Amount of cropping to apply to both width boundaries of the input activations
     */
    public static Cropping3DBuilder<?,?> builder(int cropDepth, int cropHeight, int cropWidth) {
        return innerBuilder()
                .cropping(cropDepth, cropDepth, cropHeight, cropHeight, cropWidth, cropWidth);
    }

    /**
     * @param cropLeftD Amount of cropping to apply to the left of the depth dimension
     * @param cropRightD Amount of cropping to apply to the right of the depth dimension
     * @param cropLeftH Amount of cropping to apply to the left of the height dimension
     * @param cropRightH Amount of cropping to apply to the right of the height dimension
     * @param cropLeftW Amount of cropping to apply to the left of the width dimension
     * @param cropRightW Amount of cropping to apply to the right of the width dimension
     */
    public static Cropping3DBuilder<?,?> builder(int cropLeftD, int cropRightD, int cropLeftH, int cropRightH, int cropLeftW, int cropRightW) {
        return innerBuilder()
                .cropping(cropLeftD, cropRightD, cropLeftH, cropRightH, cropLeftW, cropRightW);
    }

    /**
     * @param cropping Cropping as either a length 3 array, with values {@code [cropDepth, cropHeight, cropWidth]}, or
     * as a length 4 array, with values {@code [cropLeftDepth, cropRightDepth, cropLeftHeight, cropRightHeight,
     * cropLeftWidth, cropRightWidth]}
     */
    public static Cropping3DBuilder<?,?> builder(int[] cropping) {
        return innerBuilder()
                .cropping(cropping);
    }

public static abstract class Cropping3DBuilder<C extends Cropping3D, B extends Cropping3DBuilder<C,B>> extends
        NoParamLayerBuilder<C,B> {
        public B cropping(int ... cropping) {
            this.cropping$value = ValidationUtils.validate6NonNegative(cropping, "cropping");
            this.cropping$set = true;
            return self();
        }
}

    @Override
    public org.deeplearning4j.nn.api.Layer instantiate(NeuralNetConfiguration conf,
                                                       Collection<TrainingListener> iterationListeners, int layerIndex, INDArray layerParamsView,
                                                       boolean initializeParams, DataType networkDataType) {
        setNetConfiguration(conf);
        LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
        runInheritance();
        Cropping3DLayer ret = new Cropping3DLayer(lconf, networkDataType);
        ret.addTrainingListeners(iterationListeners);
        ret.setIndex(layerIndex);
        Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
        ret.setParamTable(paramTable);
        ret.setLayerConfiguration(lconf);
        return ret;
    }

    @Override
    public InputType getOutputType(int layerIndex, InputType inputType) {
        if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
            throw new IllegalStateException("Invalid input for 3D cropping layer (layer index = " + layerIndex
                            + ", layer name = \"" + getName() + "\"): expect CNN3D input type with size > 0. Got: "
                            + inputType);
        }
        InputType.InputTypeConvolutional3D c = (InputType.InputTypeConvolutional3D) inputType;
        return InputType.convolutional3D(c.getDepth() - cropping[0] - cropping[1],
                        c.getHeight() - cropping[2] - cropping[3], c.getWidth() - cropping[4] - cropping[5],
                        c.getChannels());
    }

    @Override
    public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
        Preconditions.checkArgument(inputType != null, "Invalid input for Cropping3D " + "layer (layer name=\""
                        + getName() + "\"): InputType is null");
        return InputTypeUtil.getPreProcessorForInputTypeCnn3DLayers(inputType, getName());
    }

    @Override
    public LayerMemoryReport getMemoryReport(InputType inputType) {
        return null;
    }



}
