/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.samediff;

import com.fasterxml.jackson.annotation.JsonIgnore;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.slf4j.Slf4j;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.params.SameDiffParamInitializer;
import org.deeplearning4j.nn.weights.WeightInit;
import org.deeplearning4j.nn.weights.WeightInitUtil;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.NetworkUtils;
import org.jetbrains.annotations.NotNull;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.factory.Nd4j;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.regularization.L1Regularization;
import org.nd4j.linalg.learning.regularization.L2Regularization;
import org.nd4j.linalg.learning.regularization.Regularization;
import org.nd4j.linalg.learning.regularization.WeightDecay;

@Slf4j
@EqualsAndHashCode(callSuper = true, doNotUseGetters = true)
@NoArgsConstructor
@SuperBuilder
public abstract class AbstractSameDiffLayer extends LayerConfiguration
    implements org.deeplearning4j.nn.api.ITraininableLayerConfiguration {

  /**
   * The regularization for the parameters (excluding biases) - for example {@link WeightDecay}
   *
   * <p>-- SETTER -- Set the regularization for the parameters (excluding biases) - for example
   * {@link WeightDecay}
   *
   * @param regularization Regularization to apply for the network parameters/weights (excluding
   *     biases)
   */
  @Getter protected List<Regularization> regularization;
  /**
   * The regularization for the biases only - for example {@link WeightDecay} -- SETTER -- Set the
   * regularization for the biases only - for example {@link WeightDecay}
   *
   * @param regularizationBias Regularization to apply for the network biases only
   */
  @Getter protected List<Regularization> regularizationBias;
  /**
   * Gradient updater. For example, {@link org.nd4j.linalg.learning.config.Adam} or {@link
   * org.nd4j.linalg.learning.config.Nesterovs}
   *
   * @param updater Updater to use
   */
  protected @Getter @Setter IUpdater updater;
  /**
   * Gradient updater configuration, for the biases only. If not set, biases will use the updater as
   * set by {@link #setUpdater(IUpdater)}
   *
   * @param biasUpdater Updater to use for bias parameters
   */
  protected @Getter @Setter IUpdater biasUpdater;

  @Getter @Setter protected GradientNormalization gradientNormalization;
  @Getter @Setter protected double gradientNormalizationThreshold = Double.NaN;
  @Getter @Setter private SDLayerParams layerParams;

  @Getter @Setter private DataType dataType;

  @Override
  public void runInheritance(@NotNull NeuralNetConfiguration conf) {
    super.runInheritance(conf);
    if (this.biasUpdater == null) this.biasUpdater = conf.getBiasUpdater();
    if (this.updater == null) this.updater = conf.getUpdater();
    if (this.regularizationBias == null || regularizationBias.isEmpty())
      this.regularizationBias = conf.getRegularizationBias();
    if (this.regularization == null || regularization.isEmpty())
      this.regularization = conf.getRegularization();
    // if( this.weightInit == null) this.weightInit = conf.getWeightInit();
    if (this.gradientNormalization == null)
      this.gradientNormalization = conf.getGradientNormalization();
    // if(this.weightInit == null) this.weightInit = conf.getWeightInit();
    if (Double.isNaN(gradientNormalizationThreshold)) {
      this.gradientNormalizationThreshold = conf.getGradientNormalizationThreshold();
    }
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    if (layerParams.isWeightParam(paramName)) {
      return regularization;
    } else if (layerParams.isBiasParam(paramName)) {
      return regularizationBias;
    }
    return null;
  }

  @JsonIgnore
  public SDLayerParams getLayerParams() {
    if (layerParams == null) {
      layerParams = new SDLayerParams();
      defineParameters(layerParams);
    }
    return layerParams;
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // Default implementation: no-op
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    // Default implementation: no-op
    return null;
  }

  /**
   * Define the parameters for the network. Use {@link SDLayerParams#addWeightParam(String,
   * long...)} and {@link SDLayerParams#addBiasParam(String, long...)}
   *
   * @param params Object used to set parameters for this layer
   */
  public abstract void defineParameters(SDLayerParams params);

  /**
   * Set the initial parameter values for this layer, if required
   *
   * @param params Parameter arrays that may be initialized
   */
  public abstract void initializeParameters(Map<String, INDArray> params);

  @Override
  public abstract org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType);

  // ==================================================================================================================

  @Override
  public ParamInitializer initializer() {
    return SameDiffParamInitializer.getInstance();
  }

  @Override
  public IUpdater getUpdaterByParam(String paramName) {
    if (biasUpdater != null && initializer().isBiasParam(this, paramName)) {
      return biasUpdater;
    } else if (initializer().isBiasParam(this, paramName)
        || initializer().isWeightParam(this, paramName)) {
      return updater;
    }
    throw new IllegalStateException("Unknown parameter key: " + paramName);
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    return false;
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return new LayerMemoryReport(); // TODO
  }

  /**
   * Returns the memory layout ('c' or 'f' order - i.e., row/column major) of the parameters. In
   * most cases, this can/should be left
   *
   * @param param Name of the parameter
   * @return Memory layout ('c' or 'f') of the parameter
   */
  public char paramReshapeOrder(String param) {
    return 'c';
  }

  protected void initWeights(int fanIn, int fanOut, WeightInit weightInit, INDArray array) {
    WeightInitUtil.initWeights(
        fanIn, fanOut, array.shape(), weightInit, null, paramReshapeOrder(null), array);
  }

  /**
   * This method generates an "all ones" mask array for use in the SameDiff model when none is
   * provided.
   *
   * @param input Input to the layer
   * @return A mask array - should be same datatype as the input (usually)
   */
  public INDArray onesMaskForInput(INDArray input) {
    if (input.rank() == 2) {
      return Nd4j.ones(input.dataType(), input.size(0), 1);
    } else if (input.rank() == 3) {
      return Nd4j.ones(
          input.dataType(),
          input.size(0),
          input.size(2)); // mask: [mb, length] vs. input [mb, nIn, length]
    } else if (input.rank() == 4) {
      // CNN style - return [mb, 1, 1, 1] for broadcast...
      return Nd4j.ones(input.dataType(), input.size(0), 1, 1, 1);
    } else if (input.rank() == 5) {
      // CNN3D style - return [mb, 1, 1, 1, 1] for broadcast...
      return Nd4j.ones(input.dataType(), input.size(0), 1, 1, 1, 1);
    } else {
      throw new IllegalStateException(
          "When using masking with rank 1 or 6+ inputs, the onesMaskForInput method must be implemented, "
              + "in order to determine the correct mask shape for this layer");
    }
  }

  public abstract static class AbstractSameDiffLayerBuilder<
          C extends AbstractSameDiffLayer, B extends AbstractSameDiffLayerBuilder<C, B>>
      extends LayerConfigurationBuilder<C, B> {

    /**
     * L1 regularization coefficient (weights only). Use {@link #l1Bias(double)} to configure the l1
     * regularization coefficient for the bias.
     */
    public B l1(double l1) {
      // Check if existing L1 exists; if so, replace it
      NetworkUtils.removeInstances(this.regularization, L1Regularization.class);
      if (l1 > 0.0) {
        this.regularization.add(new L1Regularization(l1));
      }
      return self();
    }

    /**
     * L2 regularization coefficient (weights only). Use {@link #l2Bias(double)} to configure the l2
     * regularization coefficient for the bias.<br>
     * <b>Note</b>: Generally, {@link WeightDecay} (set via {@link #weightDecay(double,boolean)}
     * should be preferred to L2 regularization. See {@link WeightDecay} javadoc for further
     * details.<br>
     */
    public B l2(double l2) {
      // Check if existing L2 exists; if so, replace it. Also remove weight decay - it doesn't make
      // sense to use both
      NetworkUtils.removeInstances(this.regularization, L2Regularization.class);
      if (l2 > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularization,
            WeightDecay.class,
            "WeightDecay regularization removed: incompatible with added L2 regularization");
        this.regularization.add(new L2Regularization(l2));
      }
      return self();
    }

    /** L1 regularization coefficient for the bias. Default: 0. See also {@link #l1(double)} */
    public B l1Bias(double l1Bias) {
      NetworkUtils.removeInstances(this.regularizationBias, L1Regularization.class);
      if (l1Bias > 0.0) {
        this.regularizationBias.add(new L1Regularization(l1Bias));
      }
      return self();
    }

    /**
     * L2 regularization coefficient for the bias. Default: 0. See also {@link #l2(double)}<br>
     * <b>Note</b>: Generally, {@link WeightDecay} (set via {@link #weightDecayBias(double,boolean)}
     * should be preferred to L2 regularization. See {@link WeightDecay} javadoc for further
     * details.<br>
     */
    public B l2Bias(double l2Bias) {
      NetworkUtils.removeInstances(this.regularizationBias, L2Regularization.class);
      if (l2Bias > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularizationBias,
            WeightDecay.class,
            "WeightDecay bias regularization removed: incompatible with added L2 regularization");
        this.regularizationBias.add(new L2Regularization(l2Bias));
      }
      return self();
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases).<br>
     * This applies weight decay <i>with</i> multiplying the learning rate - see {@link WeightDecay}
     * for more details.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient) {
      return weightDecay(coefficient, true);
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases). See {@link
     * WeightDecay} for more details.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @param applyLR Whether the learning rate should be multiplied in when performing weight decay
     *     updates. See {@link WeightDecay} for more details.
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(this.regularization, WeightDecay.class);
      if (coefficient > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularization,
            L2Regularization.class,
            "L2 regularization removed: incompatible with added WeightDecay regularization");
        this.regularization.add(new WeightDecay(coefficient, applyLR));
      }
      return self();
    }

    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details. This
     * applies weight decay <i>with</i> multiplying the learning rate.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecayBias(double, boolean)
     */
    public B weightDecayBias(double coefficient) {
      return weightDecayBias(coefficient, true);
    }

    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details<br>
     *
     * @param coefficient Weight decay regularization coefficient
     */
    public B weightDecayBias(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(this.regularizationBias, WeightDecay.class);
      if (coefficient > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularizationBias,
            L2Regularization.class,
            "L2 bias regularization removed: incompatible with added WeightDecay regularization");
        this.regularizationBias.add(new WeightDecay(coefficient, applyLR));
      }
      return self();
    }
  }
}
