/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.ocnn;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.BaseOutputLayer;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.layers.LayerValidation;
import org.deeplearning4j.nn.layers.ocnn.OCNNParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.regularization.Regularization;

@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@JsonIgnoreProperties("lossFn")
@SuperBuilder
public class OCNNOutputLayer extends BaseOutputLayer {

  /**
   * The hidden layer size for the one class neural network. Note this would be nOut on a dense
   * layer. NOut in this neural net is always set to 1 though.
   */
  @Getter private int hiddenLayerSize; // embedded hidden layer size aka "K"
  /** For nu definition see the paper */
  @Builder.Default @Getter private double nu = 0.04;
  /**
   * The number of examples to use for computing the quantile for the r value update. This value
   * should generally be the same as the number of examples in the dataset
   */
  @Builder.Default @Getter private int windowSize = 10000;
  /**
   * The initial r value to use for ocnn for definition, see the paper, note this is only active
   * when {@link #configureR} is specified as true
   */
  @Builder.Default @Getter private double initialRValue = 0.1;
  /**
   * Whether to use the specified {@link #initialRValue} or use the weight initialization with the
   * neural network for the r value
   */
  @Builder.Default @Getter private boolean configureR = true;
  /**
   * Psuedo code from keras: start_time = time.time() for epoch in range(100): # Train with each
   * example sess.run(updates, feed_dict={X: train_X,r:rvalue}) rvalue = nnScore(train_X, w_1, w_2,
   * g) with sess.as_default(): rvalue = rvalue.eval() rvalue = np.percentile(rvalue,q=100*nu)
   * print("Epoch = %d, r = %f" % (epoch + 1,rvalue))
   */
  @Builder.Default @Getter @Setter private int lastEpochSinceRUpdated = 0;

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet("OCNNOutputLayer", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();

    org.deeplearning4j.nn.layers.ocnn.OCNNOutputLayer ret =
        new org.deeplearning4j.nn.layers.ocnn.OCNNOutputLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    ret.setActivation(getActivationFn());
    if (lastEpochSinceRUpdated == 0 && configureR) {
      paramTable.get(OCNNParamInitializer.R_KEY).putScalar(0, initialRValue);
    }
    return ret;
  }

  @Override
  public long getNOut() {
    // we don't change number of outputs here
    return 1L;
  }

  @Override
  public ParamInitializer initializer() {
    return OCNNParamInitializer.getInstance();
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    // Not applicable
    return null;
  }

  public static abstract class OCNNOutputLayerBuilder<
          C extends OCNNOutputLayer, B extends OCNNOutputLayerBuilder<C, B>>
      extends BaseOutputLayerBuilder<C, B> {

    public B nOut(int nOut) {
      return nOut(Long.valueOf(nOut));
    }

    @Override
    public B nOut(long nOut) {
      throw new UnsupportedOperationException(
          "Unable to specify number of outputs with ocnn. Outputs are fixed to 1.");
    }
  }
}
